/*
 *
 * Copyright 2022 gRPC authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

package grpcutil

import (
	"testing"
)

func TestRegisteredCompressors(t *testing.T) {
	defer func(c []string) { RegisteredCompressorNames = c }(RegisteredCompressorNames)
	RegisteredCompressorNames = []string{"gzip", "snappy"}
	if got, want := RegisteredCompressors(), "gzip,snappy"; got != want {
		t.Fatalf("Unexpected compressors got:%s, want:%s", got, want)

	}
}
